# Copyright (C) Yuki Takagi 2020
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE_1_0.txt or copy at
# https://www.boost.org/LICENSE_1_0.txt)

import wave as _wave
from strcuta import cursor as _cursor
from strcuta import frq as _frq

class Type:
    def __init__(self, frq, parameter, frames):
        assert parameter.nchannels == 1
        self.frq = frq
        self.parameter = parameter
        self.frames = frames

    def __len__(self):
        return self.parameter.nframes

    def __getitem__(self, key):
        if isinstance(key, slice):
            k = _cursor.resolve_slice(key, self.parameter.framerate, self.parameter.sampwidth)
            f = self.frames[k]
            kf = _cursor.resolve_slice(key, self.frq.samplerate, 1 / self.frq.sample_interval)
            return Type(
                    frq=self.frq[kf],
                    parameter=self.parameter._replace(nframes=int(len(f) / self.parameter.sampwidth)),
                    frames=f)
        else:
            return self.frames[_cursor.resolve_to_slice(key, self.parameter.framerate, self.parameter.sampwidth)]

    def write(self, outputpath):
        with _wave.open(outputpath, mode="wb") as w:
            w.setparams(self.parameter)
            w.writeframes(self.frames)

    _audio_stream = None

    def play(self):
        from sounddevice import RawOutputStream
        Type._audio_stream = Type._audio_stream or RawOutputStream(
                channels=self.parameter.nchannels,
                dtype='int' + str(8 * self.parameter.sampwidth),
                samplerate=self.parameter.framerate * self.parameter.nchannels,)
        Type._audio_stream.start()
        Type._audio_stream.write(self.frames)

def _frq_path(wav_path):
    return wav_path[:-4] + "_wav.frq"

def load(path):
    with _wave.open(path, mode="rb") as w:
        return Type(
                frq=_frq.load(_frq_path(path)),
                parameter=w.getparams(),
                frames=w.readframes(w.getnframes())
                )

def write(wav, outputpath):
    wav.write(outputpath)



# Copyright (C) Yuki Takagi 2020
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE_1_0.txt or copy at
# https://www.boost.org/LICENSE_1_0.txt)

class Cursor:
    def _resolve(self, samplerate, samplewidth):
        pass

    def __neg__(self):
        pass

class MilliSeconds(Cursor):
    def __init__(self, value):
        self.value = value

    def __neg__(self):
        return MilliSeconds(-self.value)

    def _resolve(self, samplerate, samplewidth):
        return round(samplerate * self.value / 1000 * samplewidth)

class Samples(Cursor):
    def __init__(self, value):
        self.value = value

    def __neg__(self):
        return Samples(-self.value)

    def _resolve(self, samplerate, samplewidth):
        return round(samplewidth * self.value)

def ms(milliseconds):
    return MilliSeconds(milliseconds)

def smp(frames):
    return Samples(frames)

def resolve(cursor, samplerate, samplewidth):
    if isinstance(cursor, Cursor):
        return cursor._resolve(samplerate, samplewidth)
    elif isinstance(cursor, int):
        return round(cursor * samplewidth)
    elif cursor == None:
        return None
    else:
        raise "Unsupported Cursor type " + cursor.__class__.__name__

def resolve_slice(slice_, samplerate, samplewidth):
    return slice(
            resolve(slice_.start, samplerate, samplewidth),
            resolve(slice_.stop, samplerate, samplewidth),
            resolve(slice_.step, samplerate, samplewidth))

def resolve_to_slice(cursor, samplerate, samplewidth):
    return slice(
            resolve(cursor, samplerate, samplewidth),
            resolve(cursor, samplerate, samplewidth) + round(samplewidth))



# Copyright (C) Yuki Takagi 2020
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE_1_0.txt or copy at
# https://www.boost.org/LICENSE_1_0.txt)

from os import path as _path
from glob import glob as _glob
from itertools import chain as _chain
from collections import namedtuple as _namedtuple

_OtoNode = _namedtuple("_OtoNode", "source offset cutoff duration consonant preutterance overlap")

def load_recursive(path, encoding="cp932", node_type=_OtoNode, greedy_recursion=True):
    if greedy_recursion:
        inis = _glob(_path.join(path, "**/oto.ini"), recursive=True)
    else:
        inis = _chain(_glob(_path.join(path, "oto.ini")), _glob(_path.join(path, "*/oto.ini")))

    oto_dict = {}
    for oto_ini in inis:
        oto_dict.update(load(oto_ini, root=path, encoding=encoding, node_type=node_type))
    return oto_dict

def load(path, root=None, encoding="cp932", node_type=_OtoNode):
    oto_dict = {}
    with open(path, "r", encoding=encoding) as f:
        for line in f.readlines():
            line = line.strip()
            [source, params] = line.split("=")
            [name, offset, consonant, cutoff, preutterance, overlap] = params.split(",")

            cf = float(cutoff)
            if cf > 0:
                cf_ = cf
                duration = None
            else:
                cf_ = None
                duration = -cf

            oto_dir = _path.dirname(path) if root == None else _path.relpath(_path.dirname(path), root)
            source = _path.join(oto_dir, source)

            oto_dict[name]= node_type(
                    source=source,
                    offset=float(offset),
                    cutoff=cf_,
                    duration=duration,
                    consonant=float(consonant),
                    preutterance=float(preutterance),
                    overlap=float(overlap)
                    )
    return oto_dict

if __name__ == "__main__":
    import json

    print(json.dumps(load_recursive(".", node_type=dict), indent=2, ensure_ascii=False))



# Copyright (C) Yuki Takagi 2020
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE_1_0.txt or copy at
# https://www.boost.org/LICENSE_1_0.txt)

def load(path_, encoding="cp932"):
    prefix_map = {}
    with open(path_, encoding=encoding) as f:
        for line in f.readlines():
            prefix_map.update((line.strip().split(),))
    return prefix_map



# Copyright (C) Yuki Takagi 2020
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE_1_0.txt or copy at
# https://www.boost.org/LICENSE_1_0.txt)

from strcuta import frq
from strcuta import voicebank
from strcuta import cursor
from strcuta import prefixmap
from strcuta import wav
from strcuta import otoini



# Copyright (C) Yuki Takagi 2020
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE_1_0.txt or copy at
# https://www.boost.org/LICENSE_1_0.txt)

from os import path as _path

from strcuta import otoini as _otoini
from strcuta import prefixmap as _prefixmap
from strcuta import cursor as _cursor
from strcuta import frq as _frq
from strcuta import wav as _wav

class Cursors:
    def __init__(self, start, end_ovl, end_pre, end_con, end):
        self.start = start
        self.end_ovl = end_ovl
        self.end_pre = end_pre
        self.end_con = end_con
        self.end_fixed = self.end_con
        self.end = end

class Counts:
    def __init__(self, pre, ovl, con, full):
        self.pre = pre
        self.ovl = ovl
        self.con = con
        self.fixed = self.con
        self.full = full
        self.vow = full - con
        self.stretchable = self.vow

class Type:
    def __init__(self, rootdir, otos, prefixes):
        self.rootdir = rootdir
        self.otos = otos
        self.prefixes = prefixes

    def resolve(self, spell, key):
        return spell + self.prefixes[key]

    def oto(self, spell, key):
        return self.otos[self.resolve(spell, key)]

    def voice(self, spell, key):
        info = self.oto(spell, key)
        wav_path = _path.join(self.rootdir, info.source)
        w = _wav.load(wav_path)

        rate = w.parameter.framerate
        nframes = w.parameter.nframes

        nf_offset = _cursor.ms(info.offset)._resolve(rate, 1)
        nf_con = _cursor.ms(info.consonant)._resolve(rate, 1)
        nf_pre = _cursor.ms(info.preutterance)._resolve(rate, 1)
        nf_ovl = _cursor.ms(info.overlap)._resolve(rate, 1)
        if info.duration != None:
            nf_used = _cursor.ms(info.duration)._resolve(rate, 1)
        else:
            nf_cutoff = _cursor.ms(info.cutoff)._resolve(rate, 1)
            nf_used = nframes - nf_offset - nf_cutoff
        
        return Voice(
                _frq=_frq_memoized(
                    wav_path[:-4] + "_wav.frq",
                    nf_offset,
                    nf_used,
                    w
                    ),
                wave=w[nf_offset : nf_offset + nf_used],
                count=Counts(
                    pre=nf_pre,
                    ovl=nf_ovl,
                    con=nf_con,
                    full=nf_used,
                    )
                )

def _frq_memoized(path, offset, used, wave):
    memo = None
    def closure():
        nonlocal memo
        if memo == None:
            f = _frq.load(path, wave)
            memo = f[round(offset / f.sample_interval) : round((offset + used) / f.sample_interval)]
        return memo
    return closure


class Voice(_wav.Type):
    def __init__(self, _frq, wave, count):
        self._frq = _frq
        self.count = count
        self.cursor = Cursors(
                start=0,
                end_ovl=count.ovl,
                end_pre=count.pre,
                end_con=count.con,
                end=count.full
                )
        super().__init__(
                frq=wave.frq,
                parameter=wave.parameter,
                frames=wave.frames)

    def range_pre(self):
        return slice(self.cursor.start, self.cursor.end_pre)

    def range_ovl(self):
        return slice(self.cursor.start, self.cursor.end_ovl)

    def range_con(self):
        return slice(self.cursor.start, self.cursor.end_con)

    def range_fixed(self):
        return self.range_fixed()

    def range_vow(self):
        return slice(self.cursor.end_con, self.cursor.end)

    def range_stretchable(self):
        return self.range_vow()

    def range_intime(self):
        return slice(self.cursor.end_pre, self.cursor.end)


    def pre(self):
        return self[self.range_pre()]

    def ovl(self):
        return self[self.range_ovl()]

    def con(self):
        return self[self.range_con()]

    def fixed(self):
        return self.con()

    def vow(self):
        return self[self.range_vow()]

    def stretchable(self):
        return self.vow()

    def intime(self):
        return self[self.range_intime()]


def load(path, encoding='cp932', greedy_oto_load=False):
    otos = _otoini.load_recursive(path, encoding=encoding, greedy_recursion=greedy_oto_load)
    prefixes = _prefixmap.load(_path.join(path, 'prefix.map'), encoding=encoding)
    return Type(
            rootdir=path,
            otos=otos,
            prefixes=prefixes)



# Copyright (C) Yuki Takagi 2020
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE_1_0.txt or copy at
# https://www.boost.org/LICENSE_1_0.txt)

import wave as _wave
import struct as _struct
from strcuta import cursor as _cursor

class Type:
    def __init__(self, format_id, sample_interval, samplerate, key_frequency, comment, nsamples, frq_samples, amp_samples):
        self.format_id = format_id
        self.sample_interval = sample_interval
        self.samplerate = samplerate
        self.key_frequency = key_frequency
        self.comment = comment
        self.nsamples = nsamples
        self.frq_samples = frq_samples
        self.amp_samples = amp_samples

    def __getitem__(self, key):
        if isinstance(key, slice):
            k = _cursor.resolve_slice(key, self.samplerate, 1)
            fs = self.frq_samples[k]
            ams = self.amp_samples[k]
            return Type(
                    self.format_id,
                    self.sample_interval,
                    self.samplerate,
                    self.key_frequency,
                    self.comment,
                    len(fs),
                    fs,
                    ams)
        else:
            k = _cursor.resolve(key, self.samplerate, 1)
            return (self.frq_samples[k], self.amp_samples[k])

    def write(self, outputpath):
        buffer_ = bytearray(_fmt_hdr.size + _fmt_smp.size * self.nsamples)
        _fmt_hdr.pack_into(
                buffer_,
                0,
                self.format_id,
                self.sample_interval,
                self.key_frequency,
                self.comment,
                self.nsamples)
        for i in range(0, self.nsamples):
            _fmt_smp.pack_into(
                    buffer_,
                    _fmt_hdr.size + i * _fmt_smp.size,
                    self.frq_samples[i],
                    self.amp_samples[i])
        with open(outputpath, "wb") as f:
            f.write(buffer_)


_fmt_hdr = _struct.Struct('<8sid16si')
_fmt_smp = _struct.Struct('<dd')

def _wave_path(frq_path):
    return frq_path[:-8] + ".wav"

def load(path_, wave=None):
    with open(path_, "rb") as f:
        buffer_ = f.read()

    [format_id, sample_interval, key_frequency, comment, nsamples] = _fmt_hdr.unpack_from(buffer_)

    if wave:
        samplerate = wave.parameter.nchannels * wave.parameter.framerate / sample_interval
    else:
        with _wave.open(_wave_path(path_), "rb") as w:
            samplerate = w.getnchannels() * w.getframerate() / sample_interval

    frq_samples = []
    amp_samples = []

    for (frq, amp) in _fmt_smp.iter_unpack(buffer_[_fmt_hdr.size:]):
        frq_samples.append(frq)
        amp_samples.append(amp)

    assert len(frq_samples) == nsamples
    assert len(amp_samples) == nsamples

    return Type(
            format_id=format_id,
            sample_interval=sample_interval,
            samplerate=samplerate,
            key_frequency=key_frequency,
            comment=comment,
            nsamples=nsamples,
            frq_samples=frq_samples,
            amp_samples=amp_samples
            )

def write(frq, outputpath):
    frq.write(outputpath)



# Copyright (C) Yuki Takagi 2020
# Distributed under the Boost Software License, Version 1.0.
# (See accompanying file LICENSE_1_0.txt or copy at
# https://www.boost.org/LICENSE_1_0.txt)

from setuptools import setup
from setuptools import find_packages

from os import path
pkgdir = path.abspath(path.dirname(__file__))
with open(path.join(pkgdir, "README.md"), encoding="utf-8") as f:
    readme_md = f.read()

setup(
        name="strcuta",
        version="0.0.17rc",
        url="https://github.com/takagiy/strcuta-py",
        project_urls={
            "Repository": "https://github.com/takagiy/strcuta-py",
            "Bug reports": "https://github.com/takagiy/strcuta-py/issues"
            },
        description="Loading UTAU voice banks into the Python data structure.",
        long_description=readme_md,
        long_description_content_type="text/markdown",
        author="takagiy",
        author_email="takagiy.4dev@gmail.com",
        classifiers=[
            "License :: OSI Approved :: Boost Software License 1.0 (BSL-1.0)",
            "Development Status :: 3 - Alpha"
            ],
        packages=find_packages(),
        include_package_data=False,
        install_requires=[],
        extras_require={
            "play": ["sounddevice"]
            })
